import { ARIARoleDefinitionKey } from 'aria-query';

declare class StaticText {
    text: string | null;
    constructor(text: string | null);
    toString(): string;
}

type AsNonLandmarkRoles = 'HeaderAsNonLandmark' | 'FooterAsNonLandmark';
type VirtualRoles = 'Details' | 'DescriptionListDetails' | 'DescriptionList' | 'DescriptionListTerm' | 'EmbeddedObject' | 'PluginObject' | 'LabelText' | 'DisclosureTriangle';
type ARIARoleDefinitionKeyExtended = ARIARoleDefinitionKey | AsNonLandmarkRoles | VirtualRoles;
type A11yTreeState = {
    busy?: boolean;
    checked?: boolean;
    current?: string | boolean;
    disabled?: boolean;
    expanded?: boolean;
    pressed?: boolean;
    selected?: boolean;
};
type A11yTreeNode = {
    element: HTMLElement;
    name: string;
    role: HTMLElement['role'] | undefined;
    description: string;
    state: A11yTreeState;
    children?: (A11yTreeNode | StaticText)[];
};
type A11yTreeNodeMatch = {
    name?: string | RegExp;
    role?: HTMLElement['role'];
    description?: string | RegExp;
    state?: Partial<A11yTreeState>;
    children?: (A11yTreeNodeMatch | string | RegExp)[];
};

type MatcherOptions = {
  isNonLandmarkSubtree?: boolean;
};

declare global {
  namespace jest {
    interface Matchers<R> {
      /**
       * @description
       * Asserts that the accessibility tree for an element matches the expected structure.
       * An accessibility tree represents how a user agent (such as a screen reader) processes and communicates
       * accessibility information from the DOM. Use this matcher to check whether the pertinent accessibility
       * properties and relationships are correctly established, ensuring an accessible experience for users
       * of assistive technologies.
       *
       * The expected structure is defined by the `A11yTreeNodeMatch` interface, which can accommodate various
       * properties like role, name, description, state, and children.
       *
       * The second parameter allows for optional configuration settings
       *
       * @example
       * // Example with byRole hierarchy:
       * <nav aria-label="Main navigation">
       *   <ul>
       *     <li><a href="/home">Home</a></li>
       *     <li><a href="/about">About</a></li>
       *   </ul>
       *   <button disabled>Click me</button>
       * </nav>
       *
       * expect(screen.getByRole('navigation', { name: 'Main navigation' })).toHaveA11yTree(
       *   byRole('navigation', [
       *     byRole('list', [
       *       byRole('listitem', [
       *         byRole('link', { name: 'Home' }),
       *       ]),
       *       byRole('listitem', [
       *         byRole('link', { name: 'About' }),
       *       ]),
       *     ]),
       *     byRole('button', { name: 'Click me', disabled: true }),
       *   ])
       * );
       *
       * @see
       * - [W3C Accessibility Tree](https://www.w3.org/TR/wai-aria/#accessibility_tree)
       * - [ARIA in HTML](https://www.w3.org/TR/html-aria/)
       */
      toHaveA11yTree(match?: A11yTreeNodeMatch, options?: MatcherOptions): R;
    }
  }
}

type NameDescriptionState = Omit<A11yTreeNodeMatch, 'role' | 'children'> & A11yTreeNodeMatch['state'];
declare function byRole(role: A11yTreeNodeMatch['role'], properties: NameDescriptionState): A11yTreeNodeMatch;
declare function byRole(role: A11yTreeNodeMatch['role'], properties: NameDescriptionState, children: A11yTreeNodeMatch['children']): A11yTreeNodeMatch;
declare function byRole(role: A11yTreeNodeMatch['role'], children: A11yTreeNodeMatch['children']): A11yTreeNodeMatch;
declare function byRole(role: A11yTreeNodeMatch['role'], name: string | RegExp): A11yTreeNodeMatch;
declare function byRole(role: A11yTreeNodeMatch['role'], name: string | RegExp, children: A11yTreeNodeMatch['children']): A11yTreeNodeMatch;
declare function byRole(role: A11yTreeNodeMatch['role']): A11yTreeNodeMatch;

declare const getAccessibilityTree: (element: HTMLElement, { isNonLandmarkSubtree: userNonLandmarkSubtree, }?: MatcherOptions) => A11yTreeNode | null;

/**
 * Removes container nodes from the tree, flattening the tree.
 */
declare const pruneContainerNodes: (node: A11yTreeNode) => A11yTreeNode;

export { A11yTreeNode, A11yTreeNodeMatch, A11yTreeState, ARIARoleDefinitionKeyExtended, AsNonLandmarkRoles, MatcherOptions, VirtualRoles, byRole, pruneContainerNodes as flattenAccessibilityTree, getAccessibilityTree };
